#!perl
#
#    A little Perl program to generate a .SNP (TLIB snapshot file) from a TLIB
# journal file file (comma-separated values, usually JOURNAL.DAT).
#
# -Dave Burton   2/12/2001


# immediate output of debug prints
$| = 1;

@INC = (".","..");
# print "before 'require splitcsv.pl'\n";
require ".\\splitcsv.pl";
# print "after 'require splitcsv.pl'\n";

$verbose = 0;
if ('-w' eq $ARGV[0]) {
   $verbose = 1;
   shift @ARGV;
}

$num_args = $#ARGV+1;

if (($num_args > 3) || ($num_args < 2)) {
 print "JOUR2SNP -- Create Snapshot file from TLIB Journal file.\n" .
       "Copyright 2001, Burton Systems Software.  www.burtonsys.com, 1-919-481-0149.\n" .
       "\n" .
       "This program creates TLIB Snapshot files \"ex post facto.\"\n" .
       "\n" .
       "Usage:\n" .
       "      perl4s {option} jour2snp.pl journal.dat out.snp {level}\n" .
       "\n" .
       "or (if you have Perl v5 installed on your computer):\n" .
       "      perl {option} jour2snp.pl journal.dat out.snp {level}\n" .
       "\n" .
       "Where:\n" .
       "      {option} is optional -w to view warnings about unrecognized input lines,\n" .
       "      journal.dat is the TLIB journal file (in CSV format),\n" .
       "      out.snp is the output file to be created, and\n" .
       "      {level} is optional project level, or TRUNK, or LATEST (default TRUNK).\n" .
       "\n" .
       "To create a snapshot file for a particular date and time \"after the fact,\"\n" .
       "first copy the TLIB journal file to a temporary file, then edit the temporary\n" .
       "file to remove records of TLIB operations which were done after the desired\n" .
       "date and time, then use this program to create a snapshot file from the edited\n" .
       "temporary copy of the journal file.\n" .
       "\n" .
       "Note: Perl4s for DOS does not support long file names.\n";
       exit 1;
}

$infile = shift @ARGV;
$infile =~ tr/a-z/A-Z/;  # upper-case

$outfile = shift @ARGV;
# $outfile =~ tr/a-z/A-Z/;

if (!open(INP,$infile)) {
   printf "ERROR: could not find '%s', $!\n", $infile;
   exit 1;
}
unlink $outfile;  # unlink is just in case of buggy MARS_NWE file system that won't truncate existing files
if (!open(OUTP,">$outfile")) {
   printf "ERROR: could not create '%s', $!\n", $outfile;
   exit 1;
}

# if (3 == $num_args) {
#    $regexp = shift @ARGV;
#    print "dbg Rexexp: $regexp\n";
#    if ($regexp =~ /^(\'(?*)\'|\"(?*)\")$/) {
#       #strip quote marks
#       $regexp = $2 . $3;
#       print "dbg Rexexp: $regexp\n";
#    }
# } else {
#    $regexp = '^[0-9]#$';  # match trunk versions
# }

$levelname = 'TRUNK';
if (3 == $num_args) {
   $levelname = shift @ARGV;
   $levelname =~ tr/a-z/A-Z/;
   if ('*' eq $levelname) {
      $levelname = 'TRUNK';
   }
}

print "infile='$infile' outfile='$outfile' projlev='$levelname'\n";

$last_date_time = '?';
$numrecs = 0;
while (<INP>) {
   $numrecs++;
   if ($_ =~ /^\,\,\,\,\,/) {
      next;  # skip extended comment lines
   }
   chop;
   $_ =~ s/[\r\n]*$//;  # because 16-bit Perl lack chomp (to strip both cr and lf), we make sure here

  # if ($_ =~ /^\,\,\,\,\,\,\,\,\,\,\,"/) {
  #    # TLIB 5.0 prior to 5.00i erroneously generated only 11 leading commas
  #    # for additional comment records.  Oops.  This repairs # such records.
  #    $_ = ',' . $_;
  # }

   if ('' eq $_) {
      next;  # skip blank lines (TLIB didn't put 'em in there, though!)
   }

   # Split into fields.  This isn't quite perfect, because some older versions
   # of TLIB might have left quote marks in the comment written as the last
   # field, but since this program doesn't care about the last field anyhow,
   # that's not a problem for this program.
   @fields = &split_csv($_);
   $numfields = $#fields + 1;

   if (($numfields < 13) && (10 != $numfields)) {  # should be 10 fields (for TLIB 4.12) or 13 fields (for TLIB 5.xx)
       if ($verbose) {
          printf "Warning: line %d not in recognized format (%d fields):\n%s\n", $numrecs, $numfields, $_;
       }
   } else {
       # Okay, there are either 10 or 13 fields.
       #  (Note: an embedded quote+comma in the final (comment) field might make
       # it appear that there are 14.  Recent TLIBs won't do that, but some
       # older versions of TLIB could.)

       $v4 = ($numfields < 13);  # True for TLIB 4.12, false for TLIB 5.xx

       if ($v4) {
          # 10 fields (from old TLIB 4.12)
          ($date,$time,$cmd,$locking,$userID,$lib,$src,$lok,$vnum,$cmt) = @fields;
          $key = $src;
          $proj = '';
          $otherproj = '';
       } else {
          # 13 fields (from TLIB 5.00 or later)
          ($date,$time,$cmd,$locking,$userID,$lib,$src,$lok,$vnum,$key,$proj,$otherproj,$cmt) = @fields;
          if ('' eq $key) {
             $key = $src;
          }
       }
       if ((!defined $key) || ('' eq $key) || ('' eq $cmd)) {
          next;
       }

       $ucKey = $key;
       $ucKey =~ tr/a-z/A-Z/;

      # $tlibcmds_seen{$cmd} = 1;

       # Commands of interest are:
       #   A AF AFS AS
       #   AP APX
       #   AD ADF AX
       #   F K S (TLIB 4.12 equivalents of UF, UK and NS, respectively)
       #   N NF NS U UAF UB UBF UF UFS UFT UFW UK UKS US

       if ($v4) {
          if ('F' eq $cmd) {
             $cmd = 'UF';
          } elsif ('K' eq $cmd) {
             $cmd = 'UK';
          } elsif ('S' eq $cmd) {
             $cmd = 'NS';
          }
       }

       # all commands of interest begin with A, N, or U.
       if ($cmd !~ /^[ANU]/) {
          next;
       }

       # These are he possible U, N, and A commands (sans search mode suffixes)
       #   U,UB,UK,UBK,US,UM,UF,UBF,UKM,UKS,UFK,UBFK,UFM,UFKM,UFS,UFKS,UD
       #   N,NK,NS,NKS,NF,NFK,NFS,NFKS,
       #   A,AP,AS,AX,AD,AF,APX,ADF,AFS
       #   AU and APU are unimplemented

       $basecmd = substr($cmd,0,1);
       $suffixes = substr($cmd,1);

       # delete uninteresting suffixes:
       #    search modes (W,L,C,O,T,A,I,N)
       #    0  (no suffix)
       #    F  (fast/freshen)
       #    K  (keep checked-out/locked)
       #    U  (undo is unimplemented)
       #    S,B,M  (specify version, force branch, bump minor version)
       #    (also, there was once a beta TLIB that supported a P suffix on the U command)
       #
       # interesting suffixes that remain are:
       #   A command:  D,X (Delete or exclude from project level)
       #               P (Promote)
       #   U command:  D (discard without updating)
       #   N command:  none

       # Delete all uninteresting suffixes:
       if ('A' eq $basecmd) {
          $suffixes =~ s/[^DXP]//g;
       } elsif ('U' eq $basecmd) {
          $suffixes =~ s/[^D]//g;
       } else {
          $suffixes = '';
       }

       # commands that remain are:  A AD AP APX AX N U UD

       $cmd = $basecmd . $suffixes;

       # UD command is uninteresting;
       if ('UD' eq $cmd) {
          next;
       }

       # commands that remain are:  A AD AP APX AX N U

       $tlibcmds_seen{$cmd} = 1;

       # Get upper-case versions of project level names from this journal record
       $ucProj = $proj;
       $ucProj =~ tr/a-z/A-Z/;  # upper-case (because project level names are case-insensitive)
       $ucOtherproj = $otherproj;
       $ucOtherproj =~ tr/a-z/A-Z/;  # upper-case

       # Make a note of which project levels are mentioned in the Journal file
       if (('AP' eq $cmd) || ('APX' eq $cmd)) {
          # AP or APX command, so it affects the $otherproj level instead of $proj
          $levelsMentioned{$ucOtherproj} = 1;
       } elsif (('A' eq $cmd) || ('N' eq $cmd) || ('U' eq $cmd)) {
          if ('*' ne $ucProj) {
             $levelsMentioned{$ucProj} = 1;
          }
       }

       # If we're not looking for a particular project level, then we can
       # record either the chronologically LATEST versions or the latest TRUNK
       # versions.  Latest TRUNK is the default & more common case.
       if (('TRUNK' eq $levelname) || ('LATEST' eq $levelname)) {
          if ('A' eq $basecmd) {
             next;
          }
          if ('' eq $vnum) {
             # Uh oh!  No version number recorded!  Some older versions of TLIB
             # had a bug which could cause this to happen, unfortunately.
             $tmp = sprintf("Warning: line %d lacks version number; '%s' ignored:\n%s\n", $numrecs, $cmd, $_);
             if ($verbose) {
                print $tmp;
             }
             # This warning is important enough to display to the user ONLY if
             # this line is the last one seen for this source file.  So, we'll
             # memorize the warning message here, but discard it later if we
             # encounter a later journal record recording the version number
             # for this particular file.
             $warnings{$ucKey} = $tmp;
             next;
          }
          if (('TRUNK' eq $levelname) && ($vnum =~ /\./)) {
             # version number contains a dot, so it's not a trunk version
             next;
          }
          # U and N commands are handled identically:
          $snapshot{$ucKey} = $vnum . '|' . $key;
          if (defined $warnings{$ucKey}) {
             delete $warnings{$ucKey};
          }
          $last_date_time = $date . ',' . $time;
          next;
       }

       # ignore commands which didn't affect the desired project level
       if (('AP' eq $cmd) || ('APX' eq $cmd)) {
          # AP or APX command, so it affects the $otherproj level instead of $proj
          if ($levelname ne $ucOtherproj) {
             next;  # doesn't affect the level in which we're interested
          }
       } elsif (('A' eq $cmd) || ('N' eq $cmd) || ('U' eq $cmd)) {
          if ($levelname ne $ucProj) {
             next;  # doesn't affect the level in which we're interested
          }
       }

       if (('AD' eq $cmd) || ('AX' eq $cmd) || ('APX' eq $cmd)) {
          delete $snapshot{$ucKey};
          $last_date_time = $date . ',' . $time;
       } elsif (('A' eq $cmd) || ('AP' eq $cmd) || ('N' eq $cmd) || ('U' eq $cmd)) {
          if ('' eq $vnum) {
             $tmp = sprintf("Warning: line %d lacks version number; '%s' ignored:\n%s\n", $numrecs, $cmd, $_);
             if ($verbose) {
                print $tmp;
             }
             $warnings{$ucKey} = $tmp;
             next;
          }
          $snapshot{$ucKey} = $vnum . '|' . $key;
          if (defined $warnings{$ucKey}) {
             delete $warnings{$ucKey};
          }
          $last_date_time = $date . ',' . $time;
       } else {
          printf "ERR: line %d has unrecognized command, '%s' -- ignored:\n%s\n", $numrecs, $cmd, $_;
       }
   }
}#while

close INP;

print "$numrecs lines read from $infile.\n";

print "Commands:";
foreach $cmd (sort keys %tlibcmds_seen) {
   print " $cmd";
}
print "\n";

print "Levels mentioned:";
foreach $lvl (sort keys %levelsMentioned) {
   print " $lvl";
}
print "\n";
if (defined $levelsMentioned{'TRUNK'}) {
   print "Warning: there's an actual project level named TRUNK, but that name is special\n"
       . "to this program, so you'll have to modify this program if you need to create\n"
       . "snapshots for that particular named project level.\n";
}
if (defined $levelsMentioned{'LATEST'}) {
   print "Warning: there's an actual project level named LATEST, but that name is special\n"
       . "to this program, so you'll have to modify this program if you need to create\n"
       . "snapshots for that particular named project level.\n";
}

#       print OUTPUT '"' . $fields[0] . '","' . $fields[2] . ' ' . $fields[1] . '"' . "\n";

$cntr = 0;
foreach $src (sort keys %snapshot) {
   $cntr++;
   $tmp = $snapshot{$src};
   ($vnum,$key) = split(/\|/,$tmp);
   if ($key =~ / /) {
      $key = '"' . $key . '"';
   }
   print OUTP "$key v=$vnum\n";
}
print "$cntr source files recorded in snapshot $outfile";
if (($levelname ne '"TRUNK') && ($levelname ne 'LATEST')) {
   print " for level $levelname\n";
} else {
   print " (" . $levelname . ")\n";
}
if ($cntr) {
   print "Last relevant date/time stamp was $last_date_time\n";
}
close OUTP;

$cntr = 0;
foreach $src (sort keys %snapshot) {
   $cntr++;
   if (defined $warnings{$src}) {
      if ($src =~ / /) {
         $src = '"' . $src . '"';
      }
      print "\n" . $src . ": " . $warnings{$src};
   }
}

exit 0;

__END__

